/*
 * mqClient.h
 *
 *  Created on: Feb 22, 2017
 *      Author: vagner
 */

#pragma once

#include <mutex>
#include <string>
#include <vector>

#include <mosquitto.h>
#include <SyncVector.h>
#include <vCommon.h>

#include "mqMsg.h"

namespace vMQ
{

	//	Элемент подписки
	class subsItem
	{
	public:
		std::string Topic;	//	топик
		int qos;			//	качество обслуживания
		//
		//	ctor
		//
		subsItem ( const std::string& Topic, int qos )
			: Topic ( Topic ), qos ( qos )
		{
		}
	};

	//	Клиент mosquitto
	class mqClient : public LogItem
	{
	private:

		//	клиент MQTT
		mosquitto* mqtt;

		//	флаг очистки сессии при установлении соединения
		bool CleanSession;

		//	ClientId
		std::string ClientId;

		//	очередь исходящих
		//	Внутренняя очередь клиента
		vCommon::SyncVector out;

		//	флаг уничтожения объекта
		bool IsDisposed;

		//	семафор блокировки обработчиков событий
		std::mutex mxCallBack;

		//	семафор доступа к флагам
		std::mutex mxFlag;

		//	семафор отправки
		std::mutex mxSend;

		//	MQTT broker URL
		std::string Url;

		//	MQTT broker Port
		int Port;

		//	флаг наличия соединения
		bool IsConnected;

		//	время разрыва соединения
		i64 lost_connect_time;

		//	получить значение bool
		bool get_bool ( const bool& flag );

		//	установить значение bool
		void set_bool ( bool& flag, bool value );

		//	получить значение
		i64 get_i64 ( const i64& ref );

		//	установить значение
		void set_i64 ( i64& ref, i64 value );

		//	Отправить первое сообщение из очереди исходящих.
		void send ( );

		//
		//	<Обработчики событий>
		//
		//	on_connect
		static void on_connect ( mosquitto* mqtt, void* context, int rc );

		//	on_disconnect
		static void on_disconnect ( mosquitto* mqtt, void* context, int rc );

		//	on_message
		static void on_message ( mosquitto* mqtt, void* context, const mosquitto_message* msg );

		//	on_publish
		static void on_publish ( mosquitto* mqtt, void* context, int mid );

		//	on_subscribe
		static void on_subscribe ( mosquitto* mqtt, void* context, int mid, int count, const int* qosList );

		//	on_unsubscribe
		static void on_unsubscribe ( mosquitto* mqtt, void* context, int mid );

		//  on_log
        static void on_log ( mosquitto* mqtt, void* context, int level, const char * str );

	public:
		static const int mqEvent_Connected = 0;
		static const int mqEvent_Message = 1;
		//static const int mqEvent_Disconnected = 2;
		int keepAlive;

		//	очередь входящих
		vCommon::SyncVector* in;

		//	ctor
		mqClient ( bool CleanSession, vCommon::SyncVector* in, std::string ClientId, int keepAlive = 60 );

		//	dtor
		~mqClient ( );

		//	получить наличие соединение
		inline bool is_connected ( ) { return get_bool ( IsConnected ); }

		//	получить время потери соединения
		inline i64 get_lost_connect_time ( ) { return get_i64 ( lost_connect_time ); }

		//	connect
		bool connect ( const std::string& Url, int Port );

		//	reconnect
		bool reconnect ( );

		//	disconnect
		void disconnect ( );

		//	publish
        bool publish ( const std::string& Topic, const std::string& payload, int qos = 2, bool retain = false );
        bool publish ( mqMsg* msg );

        //	subscribe
		bool subscribe ( const std::vector<subsItem>& SubsList );

		//	unsubscribe
		bool unsubscribe ( );
		bool unsubscribe ( const std::string& Topic );
	};

} /* namespace vMQ */
